/*
 * Copyright by Vladimir N. Silyaev 1999-2000
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $vmFreeBSD: vmware/vmnet-only/freebsd/vmnet.c,v 1.18 2000/07/30 23:36:25 vsilyaev Exp $
 */

#if 0
#include <sys/param.h>
#include <sys/proc.h>
#include <sys/systm.h>
#include <sys/conf.h>
#include <sys/kernel.h>
#include <sys/malloc.h>
#include <sys/mbuf.h>
#include <sys/sockio.h>
#include <sys/socket.h>
#include <sys/filio.h>
#include <sys/poll.h>
#include <sys/uio.h>
#include <sys/vnode.h>
#include <sys/queue.h>

#include <net/if.h>
#include <net/if_arp.h>
#include <net/ethernet.h>

#include <net/bpf.h>


#ifdef BRIDGE
#include <net/bridge.h>
#endif /* BRIDGE */

#include "vmnet.h"
#include "../vm_oui.h"
#include "../vnet.h"

#define CDEV_MAJOR 201
#if defined(CDEV_MAJOR_) && CDEV_MAJOR != CDEV_MAJOR_
#error "CDEV_MAJOR != CDEV_MAJOR_"
#endif

#define	DEVICE_NAME	"vmnet"

enum vmnet_log_level {Lenter=0, Lexit=1, Linfo, Lwarning, Lfail};
#define DEBUG 0
#if DEBUG
#define DEB(x) x
#define DLog(level, fmt, args...)	printf("%s: " fmt "\n", __FUNCTION__ ,   ##args)
#else /* !DEBUG */
#define DEB(x)
#define	DLog(level, fmt, args...) 
#endif /* DEBUG */

struct vmnet_softc {
	SLIST_ENTRY(vmnet_softc) list;
	dev_t		dev; /* Back reference to device */

	struct {
		struct arpcom   arpcom;         /* interface info */
	} iface;

	struct {
	 struct selinfo	rsel;
	 struct {
		int	opened:1;
		int 	rwait:1;
	 } flags;
 	 u_int8_t ether_addr[ETHER_ADDR_LEN];
	} var;
};

SLIST_HEAD(, vmnet_softc) vmnets = SLIST_HEAD_INITIALIZER(vmnets);

static d_open_t		vmnet_open;
static d_close_t 	vmnet_close;
static d_ioctl_t	vmnet_ioctl;
static d_poll_t		vmnet_poll;
static d_read_t		vmnet_read;
static d_write_t	vmnet_write;

static int vmnet_modeevent(module_t mod, int cmd, void *arg);

static struct cdevsw vmnet_cdevsw = {
	/* open */	vmnet_open,
	/* close */	vmnet_close,
	/* read */	vmnet_read,
	/* write */	vmnet_write,
	/* ioctl */	vmnet_ioctl,
	/* poll */	vmnet_poll,
	/* mmap */	nommap,
	/* strategy */	nostrategy,
	/* name */	DEVICE_NAME,
	/* maj */	CDEV_MAJOR,
	/* dump */	nodump,
	/* psize */	nopsize,
	/* flags */	0,
	/* bmaj */	-1
};

/* 
 * Now declare the module to the system. 
 * IMPORTANT: Must be before netgraph node declaration.
 */
#ifdef MODULE_DEPEND
MODULE_DEPEND(vmnet, linux, 1, 1, 1);
#endif
DEV_MODULE(vmnet, vmnet_modeevent, 0);


static int vmnet_ifioctl(struct ifnet *, u_long, caddr_t);
static void vmnet_ifstart(struct ifnet *);
static void vmnet_ifinit(void *);

/* -=-=-=-=-=-=-=-=-= attach/detach device stuff -=-=-=-=-=-=-=-=-= */

static u_int8_t 
ether_addr[ETHER_ADDR_LEN]={0x00,0x50,0x56,0xea,0xc1,0xa8};

static int
vmnet_attach_iface(struct vmnet_softc *sc)
{
	struct ifnet *ifp = &sc->iface.arpcom.ac_if; 
	intrmask_t s;
	
	ifp->if_unit = lminor(sc->dev);
	ifp->if_name = DEVICE_NAME;
	ifp->if_mtu  = ETHERMTU;
	ifp->if_flags = IFF_BROADCAST | IFF_SIMPLEX /*| IFF_MULTICAST*/;
	ifp->if_ioctl = vmnet_ifioctl;
	ifp->if_start = vmnet_ifstart;
	ifp->if_init = vmnet_ifinit;
	ifp->if_output = ether_output;
	ifp->if_baudrate = 10000000; /* XXX */
	ifp->if_snd.ifq_maxlen = IFQ_MAXLEN;
	
	ifp->if_softc = sc;

	bcopy(ether_addr, sc->iface.arpcom.ac_enaddr, sizeof(ether_addr));
	/* Hack to unifique ethernet addr */
	sc->iface.arpcom.ac_enaddr[ETHER_ADDR_LEN-1]+=ifp->if_unit;
	DLog(Linfo, DEVICE_NAME "%d: Ethernet address: %6D", ifp->if_unit, sc->iface.arpcom.ac_enaddr, ":");

	s = splimp();
	if_attach(ifp);
	ether_ifattach(ifp);
	bpfattach(ifp, DLT_EN10MB, sizeof(struct ether_header));
	splx(s);
	
	return 0;
}

static int
vmnet_detach_iface(struct vmnet_softc *sc)
{
	struct ifnet *ifp = &sc->iface.arpcom.ac_if; 
	intrmask_t s;

	s = splimp();
	if_detach(ifp);
	splx(s);
	return 0;
}

static struct vmnet_softc *
vmnet_attach(dev_t dev)
{
	struct vmnet_softc *sc;
	int unit;

	unit = lminor(dev);
	DLog(Lenter, "%d %p", unit, dev);
	if (dev->si_drv1) {
		DLog(Lexit, "old %p, %p", dev, dev->si_drv1);
		return dev->si_drv1;
	}

  	dev = make_dev(&vmnet_cdevsw, minor(dev), UID_ROOT, GID_WHEEL, 0600, DEVICE_NAME "%d", unit); 
	if (dev==NULL)
		return (NULL);
	MALLOC(sc, struct vmnet_softc*, sizeof(*sc), M_DEVBUF, M_WAITOK);
	if (sc==NULL)
		return NULL;

	bzero(sc, sizeof(*sc));
	dev->si_drv1 = sc; /* Link together */
	sc->dev = dev;
	
	if (vmnet_attach_iface(sc)!=0) {
   		destroy_dev(dev);
		FREE(sc, M_DEVBUF);
		return NULL;	
	}


	SLIST_INSERT_HEAD(&vmnets, sc, list);
	DLog(Lexit, "new %p,%p", dev, sc);
	return sc;
}

static int
vmnet_detach(struct vmnet_softc *sc) 
{
	int error=0;

	if (sc->var.flags.opened) {
		return EBUSY;
	}
	if ( (error=vmnet_detach_iface(sc)) != 0) {
		return error;
	}
	SLIST_REMOVE_HEAD(&vmnets, list);
	destroy_dev(sc->dev);
	FREE(sc, M_DEVBUF);
	return error;
}

/* -=-=-=-=-=-=-=-=-= character device stuff -=-=-=-=-=-=-=-=-= */

int 
vmnet_open(dev_t dev, int oflag, int otyp, struct proc *p)
{
	struct vmnet_softc *sc;
	
	sc = vmnet_attach(dev);
	
	if (sc==NULL)
		return (EAGAIN);
	
	if (sc->var.flags.opened)
		return (EBUSY);
	
	bzero(&sc->var, sizeof(sc->var));

	bcopy(ether_addr, sc->var.ether_addr, sizeof(ether_addr));
	
	sc->var.flags.opened = 1;
	
	return 0;
}

int 
vmnet_close(dev_t dev, int fflag, int otyp, struct proc *p)
{
	struct vmnet_softc *sc = (struct vmnet_softc *) dev->si_drv1;
	struct mbuf *m;       
	intrmask_t s;

	/* Clear output queue */
	do {
		s = splimp();
		IF_DEQUEUE(&sc->iface.arpcom.ac_if.if_snd, m);
		splx(s);
		if (m)
			m_freem(m);
	} while(m);
	
	DLog(Linfo, "input %ld bytes, output %ld bytes", sc->iface.arpcom.ac_if.if_ibytes, sc->iface.arpcom.ac_if.if_obytes);

	sc->var.flags.opened = 0;
	return 0;
}

int 
vmnet_ioctl(dev_t dev, u_long cmd, caddr_t arg, int mode, struct proc *p)
{
	struct vmnet_softc *sc = (struct vmnet_softc *) dev->si_drv1;
	struct ifnet *ifp = &sc->iface.arpcom.ac_if; 
	int error=0;
	intrmask_t s;
	
	DLog(Lenter, "cmd 0x%04lx", cmd);
	s = splimp();
	switch (cmd) {
	case FIONBIO:
		break;
	case SIOCGIFFLAGS:
		bcopy(&ifp->if_flags, arg, sizeof(ifp->if_flags));
		break;
	case VMIO_SIOCSIFFLAGS:
		{
		 int flags=*(int *)arg;

		 flags &=0xfff; /* Drop MULTICAST and LINKx */
		 DLog(Linfo,"Set flags old 0x%04x new 0x%04x cantchange 0x%04x", ifp->if_flags, flags, IFF_CANTCHANGE);
		 ifp->if_flags = (flags & ~IFF_CANTCHANGE) | (ifp->if_flags & IFF_CANTCHANGE) | (IFF_UP);
		 DLog(Linfo,"Set flags result 0x%04x", ifp->if_flags);
		}
		break;

	case VMIO_SIOCSETMACADDR: {
	  VNet_SetMacAddrIOCTL macAddr;
	  if ((error=copyin((caddr_t)*(u_long *)arg, &macAddr, sizeof(macAddr)))!=0)
		return error;
          switch (macAddr.version) {
	  case 1:
           bcopy(macAddr.addr, sc->var.ether_addr, ETHER_ADDR_LEN);
           break;
          default:
           return EINVAL;
           break;
	  }
        }
        break;
	case OSIOCGIFADDR:
		/* Fallthrough */
	case SIOCGIFADDR:
		DLog(Linfo, "Get ethernet address: %6D", sc->var.ether_addr, ":");
		bcopy(sc->var.ether_addr, arg, ETHER_ADDR_LEN);
		/* XXX */
		break;

	case SIOCSIFADDR:
		bcopy(sc->var.ether_addr, arg, ETHER_ADDR_LEN);
		DLog(Linfo, "Set ethernet address: %6D", sc->var.ether_addr, ":");
		break;

	#define IOCTL(name) case name: DLog(Linfo, #name);break
	IOCTL(OSIOCGIFBRDADDR); 
	IOCTL(VMIO_SIOCSKEEP); 
	IOCTL(VMIO_SIOCSIFBR); 
	IOCTL(VMIO_SIOCSLADRF); 
	#undef IOCTL
	default:
		DLog(Lfail, "Unknown IOCTL 0x%04lx", cmd);
		error = EINVAL;
		break;
	}
	splx(s);

	DLog(Lenter, "result %d", error);
	return (error);
}

int 
vmnet_poll(dev_t dev, int events, struct proc *p)
{
	struct vmnet_softc *sc = (struct vmnet_softc *) dev->si_drv1;
   	int revents = 0;
	intrmask_t s;
	struct ifqueue	*xmitq_p;

	xmitq_p = &sc->iface.arpcom.ac_if.if_snd;
	s = splimp();
	if (events & (POLLIN | POLLRDNORM)) {
		if (xmitq_p->ifq_len > 0) {
			DLog(Lenter,"data arrived %d packets", xmitq_p->ifq_len);
			revents |= events & (POLLIN | POLLRDNORM);
		} else {
			selrecord(p, &sc->var.rsel);
		}
	}
	if (events & (POLLOUT | POLLWRNORM))
		revents |= events & (POLLOUT | POLLWRNORM);
	
	splx(s);
	return revents;
}

int 
vmnet_read(dev_t dev, struct uio *uio, int ioflag)
{
	struct vmnet_softc *sc = (struct vmnet_softc *) dev->si_drv1;
	struct ifnet *ifp = &sc->iface.arpcom.ac_if; 
	struct mbuf *m, *m0;
	intrmask_t s;
	int error = 0;
	int len, nread;
	struct ifqueue	*xmitq_p;

	DLog(Lenter,"");
	xmitq_p = &sc->iface.arpcom.ac_if.if_snd;
	sc->var.flags.rwait = 0;
	s = splimp();
	do {
		IF_DEQUEUE(xmitq_p, m0);
		if (m0==0) {
			if (ioflag & IO_NDELAY) {
				splx(s);
				return EWOULDBLOCK;
			}
			sc->var.flags.rwait = 1;
			DLog(Linfo,"read sleep");
			if ((error = tsleep((caddr_t)xmitq_p, PCATCH | (PZERO + 1), "vmnetread", 0)) !=0) {
				splx(s);
				return error;
			}
		}
	} while (m0==0);
	if (m0 && ifp->if_bpf) {
		bpf_mtap(ifp, m0);
	}
	ifp->if_opackets++;
	splx(s);

	len=nread=0;
	while (m0 && uio->uio_resid > 0 && error == 0) {
		len = min(uio->uio_resid, m0->m_len);
		if (len == 0) 
			break;
		error = uiomove(mtod(m0, caddr_t), len, uio);
		if (error==0) nread+=len;
		MFREE(m0, m);
		m0 = m;
	}
	if (m0) {
		ifp->if_oerrors++;
		/* Dropping mbuf */
		m_freem(m0);
	} 
	DLog(Lenter,"read %d bytes left space %d, last %d, error %d", nread, uio->uio_resid, len, error);
	return error;
}

int 
vmnet_write(dev_t dev, struct uio *uio, int ioflag)
{
	struct vmnet_softc *sc = (struct vmnet_softc *) dev->si_drv1;
	struct mbuf	*top, **mp, *m;
	intrmask_t s;
	int error = 0;
	int mlen, tlen;
	struct ifnet *ifp = &sc->iface.arpcom.ac_if; 
	struct ether_header *eh;
	


	tlen = uio->uio_resid;
	DLog(Lenter, "%d", tlen);
	if (tlen>VMNET_MRU || tlen<=0 )
		return (EIO);
	

	/* get a header mbuf */
	m=m_gethdr(M_WAIT, MT_DATA);
	if (m==NULL) {
		DLog(Lfail, "NOBUFS");
		return ENOBUFS;
	}

	mlen = MHLEN;
	
	top = 0;
	mp = &top;

	while (error == 0 && uio->uio_resid > 0) {
		m->m_len = min(mlen, uio->uio_resid);
		error = uiomove(mtod (m, caddr_t), m->m_len, uio);
		*mp = m;
		mp = &m->m_next;
		if (uio->uio_resid > 0) {
		    m = m_get (M_WAIT, MT_DATA);
		    if (m == NULL) {
			    DLog(Lfail, "NOBUFS");
			    error = ENOBUFS;
			    break;
		    }
		    mlen = MLEN;
		}
	}

	if (error) {
		DLog(Lfail, "error %d", error);
		ifp->if_ierrors++;
		if (top)
			m_freem (top);
		return error;
	}

	top->m_pkthdr.len = tlen;
	top->m_pkthdr.rcvif = ifp;
	
	ifp->if_ipackets++;

	eh = mtod(top, struct ether_header *);
	s = splimp();
	if (ifp->if_bpf) {
		bpf_mtap(ifp, top);
	}
#ifdef BRIDGE
	/* kernel Ethernet bridging */
	if (do_bridge) {
		struct ifnet	*bdg_ifp;

		bdg_ifp = bridge_in(top);
		DEB(printf("vmnet src %6D dest %6D\n", eh->ether_shost, ".", eh->ether_dhost, ".")); 
		if (BDG_DROP == bdg_ifp) { /* drop it */
			DEB(printf("BDG_DROP\n"));
			m_freem(top);
			goto done;
		}

		if (BDG_LOCAL != bdg_ifp) { /* forward it */
			DEB(printf("forward %p to %p\n", top, ifp));
			bdg_forward(&top, bdg_ifp);
		} else {
			DEB(printf("BDG_LOCAL\n"));
		}

		if (BDG_LOCAL == bdg_ifp ||
		    BDG_BCAST == bdg_ifp || BDG_MCAST == bdg_ifp) {
			goto getit;
		}

		/* drop it */
		if (top != NULL) {
			m_freem(top);
		}
		goto done;
	}

getit:

#endif /* BRIDGE */
	m_adj(top, sizeof(struct ether_header));
	ether_input(ifp, eh, top);
done:	
	splx(s);

	DLog(Lexit, "result %d", error);

	return error;
}


/* -=-=-=-=-=-=-=-=-= network interface stuff -=-=-=-=-=-=-=-=-= */

int 
vmnet_ifioctl(struct ifnet *ifp, u_long cmd, caddr_t data)
{
	int error=0;
	intrmask_t s;
	
	DLog(Lenter, "cmd 0x%04lx", cmd);
	s = splimp();

	switch(cmd) {
	case SIOCSIFFLAGS:
		break;
	case SIOCSIFADDR:
	case SIOCGIFADDR:
		error = ether_ioctl(ifp, cmd, data);
		break;
	default:
		DLog(Lfail, "Unknow ioctl 0x%04lx", cmd);
		error = EINVAL;
		break;
	}

	(void)splx(s);
	
	
	return error;
}

void 
vmnet_ifinit(void *softc)
{
	struct vmnet_softc *sc = (struct vmnet_softc *) softc;
	struct ifnet *ifp = &sc->iface.arpcom.ac_if; 
	
	
	ifp->if_flags |= IFF_RUNNING; 
	ifp->if_flags &= ~IFF_OACTIVE;
}


void 
vmnet_ifstart(struct ifnet *ifp)
{
	struct vmnet_softc *sc = (struct vmnet_softc *) ifp->if_softc;
	DLog(Lenter, "sc %p", sc);

	if (sc->var.flags.rwait) {
		sc->var.flags.rwait=0;
		wakeup(&sc->iface.arpcom.ac_if.if_snd);
	}
	selwakeup(&sc->var.rsel);
}

/* -=-=-=-=-=-=-=-=-= module load/unload stuff -=-=-=-=-=-=-=-=-= */

static int
init_module(void)
{
int error;

   	error = cdevsw_add(&vmnet_cdevsw);
	if (error) 
		return error;

	return error;
}



static int
cleanup_module(void)
{
	int error;
	struct vmnet_softc *sc;

	DLog(Lenter, "");
	while ( (sc = SLIST_FIRST(&vmnets))!=NULL ) {
		DLog(Linfo, "found %p, %p", sc->dev, sc);
		if ( (error=vmnet_detach(sc)) !=0) {
			DLog(Lfail, "%p busy", sc);
			return error;
		}
   	}
	error = cdevsw_remove(&vmnet_cdevsw);
	DLog(Linfo, "return %d", error);
	return error;
}

static int
vmnet_modeevent(module_t mod, int cmd, void *arg)
{
    int  error = 0;

    switch (cmd) {
    case MOD_LOAD:
	error = init_module();
	break;	

    case MOD_UNLOAD:
        /* fall through */
    case MOD_SHUTDOWN:
 	error = cleanup_module();
	break;

    default:	/* we only understand load/unload */
	error = EINVAL;
	break;
    }

    return (error);
}

#endif
