/*
 * Copyright by Vladimir N. Silyaev 1999-2000
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $vmFreeBSD: vmware/vmnet-only/freebsd/vmnet_linux.c,v 1.7 2000/07/31 00:36:23 vsilyaev Exp $
 */
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/sockio.h>
#include <sys/sysproto.h>
#include <sys/socket.h>
#include <sys/proc.h>
#include <sys/file.h>

#include <net/if.h>

#include <i386/linux/linux.h>
#include <i386/linux/linux_proto.h>
#include <i386/linux/linux_ioctl.h>

#define	SIOCDEVPRIVATE	0x89F0
#include "vmnet.h"
#include "../vm_oui.h"
#include "../vnet.h"

#define DEBUG 0
#if DEBUG
#define DEB(x) x
#else /* !DEBUG */
#define DEB(x)
#endif /* DEBUG */

#define LINUX_IOCTL_SET(n,low,high) \
static linux_ioctl_function_t linux_ioctl_##n; \
static struct linux_ioctl_handler n##_handler = {linux_ioctl_##n, low, high }; \
SYSINIT(n##register, SI_SUB_KLD, SI_ORDER_MIDDLE, linux_ioctl_register_handler, &n##_handler); \
SYSUNINIT(n##unregister, SI_SUB_KLD, SI_ORDER_MIDDLE, linux_ioctl_unregister_handler, &n##_handler);

LINUX_IOCTL_SET(net, 0x8900, 0x89ff);


#define	LINUX_SIOCSIFADDR 	0x8916		/* set PA address	*/
#define	LINUX_SIOCSIFFLAGS	0x8914		/* set flags		*/
#define	LINUX_SIOCSIFBR		0x8941		/* Set bridging options	*/

#define LINUX_SIOCSKEEP SIOCSKEEP 	/* not used */
#define LINUX_SIOCGKEEP SIOCGKEEP	/* not used */ 

#define LINUX_SIOCSLADRF SIOCSLADRF

#define LINUX_SIOCPORT		SIOCPORT
#define LINUX_SIOCBRIDGE	SIOCBRIDGE
#define LINUX_SIOCNETIF		SIOCNETIF
#define LINUX_SIOCSETMACADDR	SIOCSETMACADDR


static int
linux_ioctl_setmacaddr(struct proc *p,  struct linux_ioctl_args *args)
{
    struct file *fp = p->p_fd->fd_ofiles[args->fd]; 
    int error;
    VNet_SetMacAddrIOCTL macAddr;
	
    DEB(printf(__FUNCTION__ ": arg %p\n", (void*)args->arg));
    error = copyin((caddr_t)args->arg, &macAddr, sizeof(macAddr));
    DEB(printf(__FUNCTION__ ": error %d version %d\n", error, macAddr.version));
    if (error)
	    return error;

    switch (macAddr.version) {
	case 1:
	  error = fo_ioctl(fp, SIOCSIFADDR, (caddr_t)macAddr.addr, p);
	  DEB(printf("Get ethernet address: %6D\n", macAddr.addr, ":"));
	  break;
	default:
	  return EINVAL;
    }
    DEB(printf(__FUNCTION__ ": error %d\n", error));
    return error;
}
															     

static int
linux_ioctl_net(struct proc *p,  struct linux_ioctl_args *args)
{
    int error=ENOIOCTL;
    
    DEB(printf(__FUNCTION__ ": cmd 0x%04lx arg %p\n", args->cmd, (void*)args->arg));
    switch (args->cmd & 0xffff) {
#if 0	  
		case LINUX_SIOCSIFFLAGS:
		    args->cmd = SIOCSIFFLAGS;
		    error = ioctl(p, (struct ioctl_args *)args); 
		    break;
#endif		    
		case LINUX_SIOCSIFADDR:
		    args->cmd = SIOCSIFADDR;
		    error = ioctl(p, (struct ioctl_args *)args); 
		    break;
		case LINUX_SIOCSLADRF:
		case LINUX_SIOCSKEEP:
		case LINUX_SIOCSIFBR:
		    error=0;
		    break;
		case LINUX_SIOCPORT:
	    	    args->cmd = VMIO_SIOCPORT;
		    error = ioctl(p, (struct ioctl_args *)args);  
		    break;
		case LINUX_SIOCBRIDGE:
		    args->cmd = VMIO_SIOCBRIDGE;
		    error = ioctl(p, (struct ioctl_args *)args);  
		    break;
		case LINUX_SIOCNETIF:
		    args->cmd = VMIO_SIOCNETIF;
		    error = ioctl(p, (struct ioctl_args *)args);  
		    break;
		case LINUX_SIOCSETMACADDR:
		    error = linux_ioctl_setmacaddr(p, args);
		    break;
    }
    DEB(printf(__FUNCTION__ ": return %d\n", error));
    return (error);
}

LINUX_IOCTL_SET(net_sifflags, LINUX_SIOCSIFFLAGS, LINUX_SIOCSIFFLAGS);

/* What is the dumb idea to use SIOCSIFFLAGS ioctl, and directly pass flags? */
static int
linux_ioctl_net_sifflags(struct proc *p,  struct linux_ioctl_args *args)
{
	int error;

    	DEB(printf(__FUNCTION__ ": cmd 0x%04lx arg %p\n", args->cmd, (void*)args->arg));
	args->cmd = SIOCGIFFLAGS; 
	error = ioctl(p, (struct ioctl_args *)args); /* First try standart */
	if (error) { 
		args->cmd = VMIO_SIOCSIFFLAGS;  /* Than vmnet */
		error = ioctl(p, (struct ioctl_args *)args); 
	}
    	DEB(printf(__FUNCTION__ ": return %d\n", error));
	return error;
}
